import asyncio
import queue
import multiprocessing
import signal
import threading

from uvloop import _testbase as tb


class EchoServerProtocol(asyncio.Protocol):

    def connection_made(self, transport):
        transport.write(b'z')


class EchoClientProtocol(asyncio.Protocol):

    def __init__(self, loop):
        self.loop = loop

    def connection_made(self, transport):
        self.transport = transport

    def data_received(self, data):
        self.transport.close()

    def connection_lost(self, exc):
        self.loop.stop()


class FailedTestError(BaseException):
    pass


class TestIssue39Regr(tb.UVTestCase):
    """See https://github.com/MagicStack/uvloop/issues/39 for details.

    Original code to reproduce the bug is by Jim Fulton.
    """

    def run_server(self, quin, qout):
        def server_thread():
            loop = self.server_loop = self.new_loop()
            coro = loop.create_server(EchoServerProtocol, '127.0.0.1', 0)
            server = loop.run_until_complete(coro)
            addr = server.sockets[0].getsockname()
            qout.put(addr)
            loop.run_forever()
            server.close()
            loop.run_until_complete(server.wait_closed())
            try:
                loop.close()
            except Exception as exc:
                print(exc)
            qout.put('stopped')

        thread = threading.Thread(target=server_thread, daemon=True)
        thread.start()

        quin.get()
        self.server_loop.call_soon_threadsafe(self.server_loop.stop)
        thread.join(1)

    def on_alarm(self, sig, fr):
        if self.running:
            raise FailedTestError

    def run_test(self):
        try:
            for i in range(10):
                for threaded in [True, False]:
                    if threaded:
                        qin, qout = queue.Queue(), queue.Queue()
                        threading.Thread(
                            target=self.run_server,
                            args=(qin, qout),
                            daemon=True).start()
                    else:
                        qin = multiprocessing.Queue()
                        qout = multiprocessing.Queue()
                        multiprocessing.Process(
                            target=self.run_server,
                            args=(qin, qout),
                            daemon=True).start()

                    addr = qout.get()
                    loop = self.new_loop()
                    loop.create_task(
                        loop.create_connection(
                            lambda: EchoClientProtocol(loop),
                            host=addr[0], port=addr[1]))
                    loop.run_forever()
                    loop.close()
                    qin.put('stop')
                    qout.get()
        finally:
            loop.close()

    def test_issue39_regression(self):
        signal.signal(signal.SIGALRM, self.on_alarm)
        signal.alarm(5)

        try:
            self.running = True
            self.run_test()
        except FailedTestError:
            self.fail('deadlocked in libuv')
        finally:
            self.running = False
            signal.signal(signal.SIGALRM, signal.SIG_IGN)



try:
    import aiohttp
    import aiohttp.web
except ImportError:
    skip_tests = True
else:
    skip_tests = False

import asyncio
import unittest

from uvloop import _testbase as tb


class _TestAioHTTP:

    def test_aiohttp_basic_1(self):

        PAYLOAD = '<h1>It Works!</h1>' * 10000

        async def on_request(request):
            return aiohttp.web.Response(text=PAYLOAD)

        asyncio.set_event_loop(self.loop)
        app = aiohttp.web.Application()
        app.router.add_get('/', on_request)

        f = self.loop.create_server(
            app.make_handler(),
            '0.0.0.0', '0')
        srv = self.loop.run_until_complete(f)

        port = srv.sockets[0].getsockname()[1]

        async def test():
            # Make sure we're using the correct event loop.
            self.assertIs(asyncio.get_event_loop(), self.loop)

            for addr in (('localhost', port),
                         ('127.0.0.1', port)):
                async with aiohttp.ClientSession() as client:
                    async with client.get('http://{}:{}'.format(*addr)) as r:
                        self.assertEqual(r.status, 200)
                        result = await r.text()
                        self.assertEqual(result, PAYLOAD)

        self.loop.run_until_complete(test())
        self.loop.run_until_complete(app.shutdown())
        self.loop.run_until_complete(app.cleanup())

        srv.close()
        self.loop.run_until_complete(srv.wait_closed())


@unittest.skipIf(skip_tests, "no aiohttp module")
class Test_UV_AioHTTP(_TestAioHTTP, tb.UVTestCase):
    pass


@unittest.skipIf(skip_tests, "no aiohttp module")
class Test_AIO_AioHTTP(_TestAioHTTP, tb.AIOTestCase):
    pass
